from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable, Any
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DViewBox import GL3DViewBox
    from pyvol_terminal.gl_3D_graphing.widgets.GL3DViewWidget import GL3DViewWidget

import pyqtgraph.opengl as gl

from pyqtgraph.opengl import GLGraphicsItem, GLViewWidget
from pyqtgraph import opengl
from PySide6 import QtWidgets, QtCore, QtGui
import numpy as np
from OpenGL import GL
from dataclasses import dataclass, field, InitVar, asdict
import weakref
from pyqtgraph.Qt import isQObjectAlive
from pyvol_terminal.gl_3D_graphing.meta import QABCMeta, abc
import pyqtgraph as pg
from pyqtgraph import getConfigOption
from .GL3DTextItem import GL3DTextItem
from .GL3DGraphicsItems import GL3DLinePlotDataItem, GL3DTextItem
from .GL3DGraphicsItemMixin import GL3DGraphicsItemMixin
from .GL3DPlotDataItemMixin import GL3DFlatMeshPlotDataItemMixin
from .GL3DGraphicsItems import MixedGLLinePlotItem
import collections
from . import validationHelpers


import math
from pprint import pprint

from pyqtgraph.Qt import QtGui, QT_LIB
import importlib

if QT_LIB in ["PyQt5", "PySide2"]:
    QtOpenGL = QtGui
else:
    QtOpenGL = importlib.import_module(f"{QT_LIB}.QtOpenGL")



class _TickLineset(MixedGLLinePlotItem):
    def __init__(self,
                 direction: int,
                 baseCoordTicks: List[List[np.ndarray]]=None, 
                 baseCoordGrid: Tuple[np.ndarray, np.ndarray]=None,
                 borderCoordinates:  Tuple[List[float], List[float]]=None,
                 color: Tuple[float, ...]=getConfigOption('foreground'),
                 width: float=1,
                 showGrid: bool=True,
                 showBorder: bool=False,
                 parentItem=None,
                 **kwargs
                 ):
        
        mode = "lines"    # Keep mode lines for optimized painting for grouped line items 
        
        super().__init__(parentItem=parentItem, color=color, width=width, mode=mode, **kwargs)
        self.direction=direction
        self.baseCoordTicks=baseCoordTicks
        self.baseCoordGrid=baseCoordGrid
        self.tickVector=None
        self.gapOffset=None
        self._border_flag=showBorder
        self.border_cells=None
        self.showGrid=showGrid
        self.base_tick_coords: List[List[float]]=None
        self.base_grid_coords: List[List[float]]=None
        self.endGridCoord=0
        
        if self._border_flag:
            self.border_cells = borderCoordinates


    def _paintHelper(self):
        return super()._paintHelper()
    
    def _childPaint(self):
        return opengl.GLLinePlotItem.paint(self)

    
    def childrenBoundingRect(self):...

    def updateTickPositions(self, tickLevelsWorld):
        tick_arr = []
        t_T=None
        tick_pos=None

        for jdx, tick_coords in enumerate(tickLevelsWorld):
            for tick_pos in tick_coords[1]:
                base_tick_offsets = self.baseCoordTicks[jdx].copy()
                t_0 = base_tick_offsets[0].copy()
                t_T = base_tick_offsets[1].copy()
                t_0[self.direction] = tick_pos
                t_T[self.direction] = tick_pos
                
                tick_arr.append(t_0)
                tick_arr.append(t_T)
                
                if self.showGrid:
                    g_0 = self.baseCoordGrid[0].copy()
                    g_T = self.baseCoordGrid[1].copy()
                    g_0[self.direction] = tick_pos
                    g_T[self.direction] = tick_pos
                    tick_arr.append(g_0)
                    tick_arr.append(g_T)
                    
        if self._border_flag:
            tick_arr.append(self.border_cells[0]) 
            tick_arr.append(self.border_cells[1]) 
        

        tick_arr = np.array(tick_arr)   
        self.setData(pos=tick_arr)


class GLAxisTextItem(GL3DTextItem):
  
    def __init__(self,
                 tickLevel: float=None,
                 localPosition: np.ndarray=None,
                 #orientation:int=None, 
                 #offset=None,
                 ref: int=None,
                 parentItem=None, 
                 **kwargs
                 ):        
        
        super().__init__(parentItem=parentItem, **kwargs)
        self._idRef=ref
        self.localPosition=localPosition
        
        self.setTickLevel(tickLevel)
        #self.setOffset(offset)
        
        
    def setTickLevel(self, tickLevel):self._tickLevel=tickLevel
    def setOffset(self, offset):self._offset=offset
    
    def tickLevel(self):return self._tickLevel
    def offset(self):return self._offset
    def idRef(self):return self._idRef
    
    def setPos(selfm, kwargs):
        return kwargs
    
    def scale(self, *args, **kwargs):
        print("scale")
        return super().scale(*args, **kwargs)   
    
    
    def setData1(self, **kwargs):
        if "pos" in kwargs:
            kwargs = self.setPos(kwargs)
        return super().setData(**kwargs)

    
    @QtCore.Slot(object)
    def updateTextVisibilityForItem(self, axis_item: GLAxisItem):
        if self.idRef() is None:
            return
        if self.idRef() > axis_item._currentTicks[self._tickLevel]-1:
            self.hide()
        else:
            self.show()
            

@dataclass(slots=True)
class AxisGeometry:
    axis: int
    origin: np.ndarray[float]
    planeRectSize: np.ndarray[float]
    graphPlaneOffset: np.ndarray[float] | float = field(default=0.) 
    tickLength: float = field(default=0.07)
    planeTickOffset: float = field(default=0.) 
    tickTextOffset: float = field(default=0.)
    tickTextMinWidth: float = field(default=0.25)
    labelOffset: float  = field(default=0.)
    
    planeAxis: int = field(init=False, default=None)
    _Orientation: int = field(init=False, default=False)
    _cachedData: Dict[str, float|np.ndarray] = field(default=dict)
    _bounds: List[Tuple[float, float]] = field(default=list)
    
    def __post_init__(self):
        self._cachedData = {"planeBoundingRect" : None,
                            "tickBoundingRect" : None,
                            "tickTextBoundingRect" : None,
                            "labelPosition" : None,
                            "tickTextPosition" : None}
        
        self.planeAxis = [i for i, val in enumerate(self.planeRectSize) if val != 0 and i != self.axis].pop(0)
        self._Orientation = int(-1 * self.planeRectSize[self.planeAxis] / abs(self.planeRectSize[self.planeAxis]))
        
        if isinstance(self.graphPlaneOffset, float):
            _perpendicular_plane_axis = [i for i in range(3) if i != self.axis and i != self.planeAxis].pop(0)
            planeOffset_new = np.zeros(3)
            planeOffset_new[_perpendicular_plane_axis] = self.graphPlaneOffset
            self.graphPlaneOffset = planeOffset_new
        
        self.planeBoundingRect()
        self.tickBoundingRect()
        self.labelPosition()
        self.getRangeNoLabel()
    
    def Orientation(self): return self._Orientation
    
    def originPosition(self) -> np.ndarray:
        return self.origin
    
    def planePosition(self) -> np.ndarray:
        parent_position = self.originPosition().copy()
        return parent_position + self.graphPlaneOffset
    
    def planeBoundingRect(self) -> np.ndarray:
        if not self._cachedData["planeBoundingRect"] is None:
            return self._cachedData["planeBoundingRect"]
        else:
            self._cachedData["planeBoundingRect"]=np.append(self.planePosition(), self.planeRectSize)
            return self._cachedData["planeBoundingRect"]
        
    def tickBoundingRect(self) -> np.ndarray:
        if not self._cachedData["tickBoundingRect"] is None:
            return self._cachedData["tickBoundingRect"]
        planeBoundingRect = self.planePosition().copy()
        position = planeBoundingRect[:3]
        position[self.planeAxis] += self.compute_offset(1 * self.planeTickOffset)

        size = [0., 0., 0.]
        size[self.axis] = self.planeRectSize[self.axis]
        size[self.planeAxis] = self.compute_offset(self.tickLength)            
        
        self._cachedData["tickBoundingRect"]=np.append(position, size)
        return self._cachedData["tickBoundingRect"]
    
    def tickTextPosition(self):
        if not self._cachedData["tickTextPosition"] is None:
            return self._cachedData["tickTextPosition"]
        
        boundingRect = self.tickBoundingRect().copy()
        boundingRect[self.planeAxis] += boundingRect[3 + self.planeAxis] + self.compute_offset(self.tickTextOffset)
        self._cachedData["tickTextPosition"] = boundingRect[:3]
        return self._cachedData["tickTextPosition"]
    
    def tickTextBoundingRect(self):
        if not self._cachedData["tickTextBoundingRect"] is None:
            return self._cachedData["tickTextBoundingRect"]
        
        size = [0., 0., 0.]
        size[self.axis] = self.planeRectSize[self.axis]
        size[self.planeAxis] = self.compute_offset(self.tickTextMinWidth)
        self._cachedData["tickTextBoundingRect"] = np.append(self.tickTextPosition(), size)
        return self._cachedData["tickTextBoundingRect"]
        
    def labelPosition(self):
        if not self._cachedData["labelPosition"] is None:
            return self._cachedData["labelPosition"]

        parentBoundingRect = self.tickTextBoundingRect()
        position = parentBoundingRect.copy()[:3]
        position[self.axis] += 0.5 * self.planeRectSize[self.axis]
        position[self.planeAxis] += self.compute_offset(self.labelOffset)
        
        self.getRangeNoLabel()
        
        if any(lo <= val <= hi for val, (lo, hi) in zip(position, self._bounds)):
            lim_idx = (self.Orientation() + 1) // 2
            lim = self._bounds[self.planeAxis][lim_idx]
            position = parentBoundingRect.copy()[:3]
            position[self.axis] += 0.5 * self.planeRectSize[self.axis]
            position[self.planeAxis] = lim + self.compute_offset(self.labelOffset)
        self._cachedData["labelPosition"]=position
        return position
    
    def compute_offset(self, offset):
        return self.Orientation() * offset
    
    def getRangeNoLabel(self):
        axis_data_keys = ["planeBoundingRect", "tickBoundingRect", "tickTextBoundingRect"]
        
        lims = [[],[],[]]
        
        for key in axis_data_keys:
            boundingRect = self._cachedData[key]
            for i in range(3):
                lims[i].append(boundingRect[i])
                lims[i].append(boundingRect[i] + boundingRect[3 + i])
        
        self._bounds = [(min(data), max(data)) for data in lims]



class GLAxisItem(GL3DGraphicsItemMixin):
    sigAxisChanged = QtCore.Signal(QtCore.QObject)
    
    
    """
    
    axis indexes are ordered x == 0, y == 1, z == 2
       
    """
    
    def __init__(self,
                 direction: int,
                 origin: Tuple[int, int, int],
                 planeRectSize: Tuple[int, int, int],                 
                 text: str="",
                 labelFont: QtGui.QFont | None=QtGui.QFont("Neue Haas Grotesk", 16),
                 labelColor: str=getConfigOption('foreground'),
                 labelOffset: float=0., 
                 labelResizeToWorld: bool=True,
                 labelAutoUpdateAnchor: bool=None,
                 labelAnchor: Tuple[float, float]=(0.5, 0.5),
                 showLabel: bool=None,
                 valuesFont: QtGui.QFont | None=QtGui.QFont("Neue Haas Grotesk", 10),
                 valuesColor: Tuple[float,...]=getConfigOption('foreground'),
                 tickLength: float=None,
                 ValuesOffset: float=0.01,
                 tickColor: Tuple[float,...]=getConfigOption('foreground'),
                 linkView: GL3DViewBox=None,
                 parentItem=None,
                 maxTickLength=-5,
                 showPlaneBorder: bool=True,
                 showValues: bool=True,
                 planeTickOffset: float=0.,
                 tickTextOffset: float=0.,
                 tickBorderOffset: float=0.1,
                 graphPlaneOffset: float=0.,
                 invertedFromWorld: bool=False,
                 valuesAnchor: Tuple[float, float]=(0., 0.),
                 textResizeToWorld: bool=True,
                 textAutoUpdateAnchor: bool=True,
                 focus: bool=False,
                 showGrid: bool=True,
                 **kwargs
                 ):
        super().__init__(parentItem=parentItem)
        self.flipped=False
        
        self.setFlag(self.GLGraphicsItemFlag.ItemHasNoContents)
        self.setParentItem(parentItem)

        self._nTicksMajor = 6
        
        self._linkedView = None
        self._name=None
        if tickLength is None:
            tickLength = abs([size for ax, size in enumerate(planeRectSize) if ax != direction and size != 0].pop(0)) / 15
        
        self.style = {
            #        'tickTextOffset': [5, 2],  ## (horizontal, vertical) spacing between text and axis
                    'tickTextWidth': 30,  ## space reserved for tick text
                    'tickTextHeight': 18,
                    'autoExpandTextSpace': True,  ## automatically expand text space if needed
                    'autoReduceTextSpace': True,
                #   'hideOverlappingLabels': hide_overlapping_labels,
                    'valuesFont': valuesFont,
                    "valuesColor" : valuesColor,
                    "tickColor" : tickColor,
                    'stopAxisAtTick': (False, False),  ## whether axis is drawn to edge of box or to last tick
                    'textFillLimits': [  ## how much of the axis to fill up with tick text, maximally.
                        (0, 0.8),    ## never fill more than 80% of the axis
                        (2, 0.6),    ## If we already have 2 ticks with text, fill no more than 60% of the axis
                        (4, 0.4),    ## If we already have 4 ticks with text, fill no more than 40% of the axis
                        (6, 0.2),    ## If we already have 6 ticks with text, fill no more than 20% of the axis
                                    ],
                    'showValues': showValues,
                    "showPlaneBorder"  : showPlaneBorder,
                    "tickLength" : tickLength,
                    "maxNticksperLevel" : 10,
                    "tickWidth" : [2.5, 1.5],
                    'maxTickLevel': 1,
                    'maxTextLevel': 1,
                    'tickAlpha': None,  ## If not none, use this alpha for all ticks.
                    "invertedFromWorld" : int(invertedFromWorld),
                    "valuesAnchor" : valuesAnchor,
                    "textAutoUpdateAnchor" : textAutoUpdateAnchor,
                    "textAnchorVar" : None,
                    "tickTextOffset" : tickTextOffset,
                    "showGrid" : showGrid,
                    "focus" : focus
                }
        self.textWidth = 30  ## Keeps track of maximum width / height of tick text
        self.textHeight = 18
        self._currentTicks=[0 for _ in range(self.style["maxTickLevel"] + 1)]
        
        self.transformCondtions = {}
        
        self.offsetStyle = {"graphPlaneOffset" : graphPlaneOffset,
                            "tickBorder" : tickBorderOffset,
                            "ticks" : planeTickOffset,
                           # "tickText" : ValuesOffset,
                            "labelOffsetPlane" : labelOffset,
                            }
    
        self.fixedWidth = None
        self.fixedHeight = None

        self.logMode = False

        self._tickDensity = 1.0   # used to adjust scale the number of automatically generated ticks
        self._tickLevels  = None  # used to override the automatic ticking system with explicit ticks
        self._tickSpacing = None  # used to override default tickSpacing method
        self.scale = 1.0
        self.autoSIPrefix = True
        self.autoSIPrefixScale = 1.0

        self.labelText = ""
        self.labelUnits = ""
        self.labelUnitPrefix = ""
        self._labelItem = None
        self._GLItems = {"values" : None,
                         "ticks" : None, 
                         "tickLineset" : None,
                         "label" : None
                         }
        self._GLItems: Dict[str, GL3DTextItem|GL3DLinePlotDataItem]
        
        self.tickStyle = {"textResizeToWorld" : textResizeToWorld,
                          "valuesAnchor" : valuesAnchor,
                          "textAutoUpdateAnchor" : textAutoUpdateAnchor or valuesAnchor != (0, 0),
                          "showValues" : showValues,
                          'valuesFont': valuesFont,
                          "tickLength" : tickLength,
                          }
        self._siPrefixEnableRanges = None
        
        
        
        
        
        showLabel = False
        
        self._axisGeometry = AxisGeometry(axis=direction,
                                           origin=np.array(origin),
                                           planeRectSize=np.array(planeRectSize),
                                           graphPlaneOffset=graphPlaneOffset,
                                           tickLength=tickLength,
                                 #          planeTickOffset=planeTickOffset,
                                           tickTextOffset=tickTextOffset,
                                           labelOffset=labelOffset,
                                           )        
        showLabel = showLabel if not bool(text) is None else showValues
        labelAutoUpdateAnchor = showLabel and (labelAutoUpdateAnchor or labelAnchor != (0, 0))
        self.labelStyle = {"labelText" : text,
                           "labelColor" : labelColor or getConfigOption('foreground'),
                           "labelFont" : labelFont,
                           "labelAnchor" : labelAnchor,
                           "labelAutoUpdateAnchor" : labelAutoUpdateAnchor,
                           "labelResizeToWorld" : labelResizeToWorld,
                           "labelOffsetPlane" : labelOffset,
                           "showLabel" : showLabel,
                           }
        
        self._baseTickWorldCoords = None
        


        self._GLItems["planeBorder"] = self._initPlaneRectBorderItem()
        self._GLItems["text"] = self._initValueTextItems()
        self._GLItems["label"] = self._initLabelTextItem()
        self._GLItems["tickLineset"]= self._initTickItems()
        
        self.setRange(self._axisGeometry.originPosition()[self.axisDirection()], self._axisGeometry.originPosition()[self.axisDirection()] + self._axisGeometry.planeRectSize[self.axisDirection()])
        self.setLabel(**self.labelStyle)

        self._linkedView = None
        
        if linkView is not None:
            self._linkToView_internal(linkView)
            
        self.setTickStyle(valuesColor=valuesColor,
                          valuesFont=QtGui.QFont("Neue Haas Grotesk", 10),
                          valuesAnchor=self.style["valuesAnchor"],
                          )
        
        
        self._buildOptimizers()
        
        print(f"origin: {self._axisGeometry.originPosition()}")
        if direction == 2 and planeRectSize[1] != 0:
            print(planeRectSize)
            import sys 
            sys.exit()
            
        
        
    def _setViewBoxRange(self, viewBoxRange):
        self.viewBoxRange=viewBoxRange
        
    def _setDataHelper(self):...
    def clipDataFromVRange(self):...
    
    @classmethod
    def fromViewBoxRange(cls,
                         viewRange: list[list[float]] | Dict[str, list[float]],   # limits of the viewBox in world coordinates [[x_min, x_max], [y_min, y_max], [z_min, z_max]] or {"x" : [x_min, x_max], "y" : [y_min, y_max], "z" : [z_min, z_max]}
                         direction: str | int,
                         origin: Tuple[int, int, int]|Dict[str, int]=None, #    extremum values must be either 0 or 1, and orients the axis origin in world coordinates. {"x" : int, "y" : int ,..} or (int, int ,int)
                         planeDirection: str | int | None = None,         # planeDirection is the direction that shares the rectangular-plane with the axis-direction along the viewRange. e.g. for the unit-cube with direction==x planeDirection==y for the xy plane
                         **kwargs
                         ) -> 'GLAxisItem':
        planeDirection = validationHelpers.verifyPlaneDirection(direction, planeDirection)
        direction_size = [d - origin[direction] for d in viewRange[direction] if d != origin[direction]][0]
        planeDirection_size = [d - origin[planeDirection] for d in viewRange[planeDirection] if d != origin[planeDirection]][0]
        
        planeRectSize = [0, 0, 0]
        planeRectSize[direction] = direction_size
        planeRectSize[planeDirection] = planeDirection_size

        axisItem  = cls(direction=direction,
                        origin=origin,
                        planeRectSize=planeRectSize,
                        
                        planeDirection=planeDirection,
                        **kwargs)    
        return axisItem
    
    def _initPlaneRectBorderItem(self) -> opengl.GLLinePlotItem:
        planeBoundingRect = self._axisGeometry.planeBoundingRect()
        
        p1 = planeBoundingRect[:3]
        p2 = p1.copy()
        p2[self.axisDirection()] += planeBoundingRect[3 + self.axisDirection()]
        
        p3 = p1.copy()
        p3 += planeBoundingRect[3:]
        
        p4 = p1.copy()
        p4[self._axisGeometry.planeAxis] += planeBoundingRect[3 + self._axisGeometry.planeAxis]
        
        segments = [p1, p2, p2, p3, p3, p4, p4, p1]
        
        borderline = MixedGLLinePlotItem(pos=segments,
                                            color= self.style["tickColor"],
                                            glOptions="translucent",
                                            mode="lines",
                                            )
        borderline.setParentItem(self)
        return borderline
    

    

    def _validate_kwargs(self, direction, orthoCoords, perpendicular_axis, viewBoxRange):

        c_viewBoxRange = viewBoxRange.copy() if isinstance(viewBoxRange, list) else viewBoxRange.tolist().copy()
        _ = c_viewBoxRange.pop(direction)
        orthoCoords = list(orthoCoords)
        if not all(orthoCoords[idx] in ax for idx, ax in enumerate(c_viewBoxRange)):
            raise ValueError("Orthogonal Coordinates do not match the viewBox range")



    def _childPaint(self):...
    def _paintHelper(self):... 

      
    def _initValueTextItems(self) -> List[List[GLAxisTextItem]]:
        gl_objects_level = []
        idx=0
        
        tickTextPosition = self._axisGeometry.tickTextPosition() + self.style["tickTextOffset"]
        for jdx in range(self.style["maxTextLevel"] + 1):
            gl_objects = []
            for idx in range(self.style["maxNticksperLevel"]):
                localPosition = tickTextPosition.copy()
                gl_object = GLAxisTextItem(tickLevel=jdx,
                                           localPosition=localPosition,
                                           ref=idx,
                                           anchor=self.tickStyle["valuesAnchor"],
                                           font=self.tickStyle["valuesFont"], 
                                           resizeToWorld=True
                                           )
                gl_object.setParentItem(self)
                gl_object.hide()
                self.sigAxisChanged.connect(gl_object.updateTextVisibilityForItem)
                gl_objects.append(gl_object)
                idx+=1
            gl_objects_level.append(gl_objects)
        return gl_objects_level





    def axisDirection(self):
        return self._axisGeometry.axis
                
    
    def tickText(self, minVal:float, maxVal:float, size: float):
        minVal, maxVal = sorted((minVal, maxVal))

        minVal *= self.scale
        maxVal *= self.scale

        ticks = []
        ticksWorld = []
        
        
        tickLevels = self.tickSpacing(minVal, maxVal, size)
        allValues = np.array([])
        for i in range(len(tickLevels)):
            spacing, offset = tickLevels[i]

            ## determine starting tick
            start = (math.ceil((minVal-offset) / spacing) * spacing) + offset

            ## determine number of ticks
            num = int((maxVal-start) / spacing) + 1
            values = (np.arange(num) * spacing + start) / self.scale
            ## remove any ticks that were present in higher levels
            ## we assume here that if the difference between a tick value and a previously seen tick value
            ## is less than spacing/100, then they are 'equal' and we can ignore the new tick.
            close = np.any(
                np.isclose(
                    allValues,
                    values[:, np.newaxis],
                    rtol=0,
                    atol=spacing/self.scale*0.01
                ),
                axis=-1
            )
            values = values[~close]
            allValues = np.concatenate([allValues, values])
            ticks.append((spacing/self.scale, values.tolist()))
            ticksWorld.append((self.getWorldCoord(spacing/self.scale), self.getWorldCoord(values).tolist()))
            self._currentTicks[i] = len(values.tolist())
        return ticks, ticksWorld

    def updateText(self, tickLevels, tickLevelsWorld):
        
        for jdx, (ticks, ticksWorld) in enumerate(zip(tickLevels, tickLevelsWorld)):
            for idx, (tick, tickWorld) in enumerate(zip(ticks[1], ticksWorld[1])):
                text_item=self._GLItems["text"][jdx][idx]
                text_item.localPosition[self.axisDirection()] = tickWorld
                text_item.blockUpdates(True)
                text_item.setData(pos=text_item.localPosition, text=str(round(tick,2)))
                text_item.blockUpdates(False)
                

  
    def setRange(self, mn: float, mx: float):
        self.range = [mn, mx]
        tickLevels, tickLevelsWorld = self.tickText(*self.range, 1)
        if self.style["showValues"]:
            self.updateText(tickLevels, tickLevelsWorld)
        self._GLItems["tickLineset"].updateTickPositions(tickLevelsWorld)
        self.update()
        
    def setTextFont(self, font=None):
        if font is None:
            getConfigOption("foreground")
            
    def setLabel(
                self,
                labelText: str | None=None,
                labelColor: str | None=None,
                labelFont: str | None=None,
                labelAnchor: str=None,
                labelResizeToWorld: bool=None,
                labelAutoUpdateAnchor: bool=None,
                showLabel: bool=None,
                **kwargs):  
        
        
        if not labelText is None:
            
            
            self.labelStyle["labelText"] = labelText
        
        visible = bool(self.labelStyle["labelText"])
        if visible:
            self._GLItems["label"].setData(text=self.labelStyle["labelText"])
        else:
            self._GLItems["label"].setData(text="")
        if not showLabel is None:
            if showLabel and not visible:
                raise
            
        self.showLabel(visible)
        
        del kwargs["labelOffsetPlane"]

        if not labelColor is None:
            self.labelStyle["labelColor"] = labelColor
        if not labelFont is None:
            self.labelStyle["labelFont"] = labelFont
        if not labelAnchor is None:
            self.labelStyle["labelAnchor"] = labelAnchor
        if not labelResizeToWorld is None:
            self.labelStyle["labelResizeToWorld"] = labelResizeToWorld and visible
            self.labelStyle["labelText"]
        
        if not labelAutoUpdateAnchor is None:
            self.labelStyle["labelAutoUpdateAnchor"] = (labelAutoUpdateAnchor or labelAnchor != (0, 0)) and visible
        
        kwargs = {"resizeToWorld" : self.labelStyle["labelResizeToWorld"],
                  "text" : self.labelStyle["labelText"],
                  "color" : self.labelStyle["labelColor"],
                  "font" : self.labelStyle["labelFont"],
             #     "anchor" : self.labelStyle["labelAnchor"],
                  }
        
        self._GLItems["label"].setData(**kwargs)
    
    def originPosition(self):
        return self._axisGeometry.originPosition()
    
    def planeAxes(self):
        axes = [self._axisGeometry.axis, self._axisGeometry.planeAxis]
        axes.sort()
        return axes
    
    def showLabel(self, show: bool=True):
        if show:
            self._GLItems["label"].show()
        else:
            self._GLItems["label"].hide()
    
    def setTickStyle(self,
                     tickColor: str | None=None,
                     tickAnchor: Tuple[float, float]=(0., 0.),
                     valuesFont: str | None=None,
                     textResizeToWorld: bool=True,
                     textAutoUpdateAnchor: bool=None,
                     showValues: bool=None,
                     *args,
                     **kwargs
                     ):
        
        if not tickColor is None:
            self.tickStyle["tickColor"] = tickColor            
        if not valuesFont is None:
            self.tickStyle["valuesFont"] = valuesFont            
        if not tickAnchor is None:
            self.tickStyle["tickAnchor"] = tickAnchor
        if not textResizeToWorld is None:
            self.tickStyle["textResizeToWorld"] = textResizeToWorld
        if not tickAnchor is None:
            self.tickStyle["tickAnchor"] = tickAnchor
        if not showValues is None:
            self.tickStyle["showValues"] = showValues
        
        if not textAutoUpdateAnchor is None:
            self.tickStyle["textAutoUpdateAnchor"] = (textAutoUpdateAnchor or tickAnchor != (0, 0)) and self.tickStyle["showValues"]

        if self.tickStyle["showValues"]:
        
            if "values" in self._GLItems and valuesFont:
                for tick_level in self._GLItems["text"]:
                    for item in tick_level:
                        item.setData(font=self.tickStyle["valuesFont"])
    
    def _initLabelTextItem(self) -> GLAxisTextItem:
        
        pos = np.zeros(3)
        pos[self.axisDirection()] = 1
        orientation = self._axisGeometry.Orientation()
        pos[self._axisGeometry.planeAxis] = 0.75 * self._axisGeometry.Orientation()
        labeItem = GLAxisTextItem(localPosition=self._axisGeometry.labelPosition(),
                                  pos=pos,#self._axisGeometry.labelPosition(),
                                  anchor=self.labelStyle["labelAnchor"],
                                  font=self.labelStyle["labelFont"], 
                                  )
        labeItem.setParentItem(self)
        return labeItem
        
    

    def _initTickItems(self) -> _TickLineset:
        
        
        ticksBoundingRect = self._axisGeometry.tickBoundingRect()
        
        
        tick_pos_container = []
        
        
        
        for idx in range(self.style["maxTickLevel"] + 1):
            
            tickStart = ticksBoundingRect[:3]
            tickEnd = tickStart.copy() + ticksBoundingRect[3:] / (1+idx)
            
            
            tick_coords = [tickStart, tickEnd]
            tick_pos_container.append(tick_coords)
            

        gridStart = self._axisGeometry.originPosition().copy()        
        
        
        
        axisBoundingRect = self._axisGeometry.planeBoundingRect()
        
        gridStart = axisBoundingRect[:3]
        
        gridEnd = axisBoundingRect[:3] + axisBoundingRect[3:]
            
        
        
        gridPositions = (gridStart.tolist().copy(), gridEnd.tolist().copy())

        
        
        start = self._axisGeometry.tickBoundingRect()[:3]
        end = start.copy()
        end[self.axisDirection()] = self._axisGeometry.tickBoundingRect()[3 + self.axisDirection()]

        
        borderPositions = [start.tolist(), end.tolist()]
        

        tickItem = _TickLineset(self.axisDirection(),
                               baseCoordTicks=tick_pos_container,
                               baseCoordGrid=gridPositions,
                               borderCoordinates=borderPositions,
                               glOptions="translucent",
                               antialias=True,
                               color=self.style["tickColor"],
                               width=self.style["tickWidth"][0],
                               showGrid=self.style["showGrid"],
                               showBorder=self.style["showPlaneBorder"],
                               )
        tickItem.setParentItem(self)
        return tickItem
        
    def checkWidths(self, textNum: str):...
    
    @QtCore.Slot(object)
    @QtCore.Slot(object, object)
    def linkedViewChanged(self, view: GL3DViewBox, newRange=None):
        if newRange is None:
            newRange = view.viewRange()[self.axisDirection()]
        self.setRange(*newRange)

    def tickSpacing(self, minVal: float, maxVal: float, size: float):
        _tickDensity=1
        dif = abs(maxVal - minVal)
        if dif == 0:
            return []

        ref_size = 300. # axes longer than this display more than the minimum number of major ticks
        minNumberOfIntervals = max(
            2.25,       # 2.0 ensures two tick marks. Fudged increase to 2.25 allows room for tick labels.
            2.25 * _tickDensity * math.sqrt(size/ref_size) # sub-linear growth of tick spacing with size
        )

        majorMaxSpacing = dif / minNumberOfIntervals

        mantissa, exp2 = math.frexp(majorMaxSpacing)
        p10unit = 10. ** (
            math.floor(
                (exp2-1)
                / 3.32192809488736
            ) - 1
        )
        
        if 100. * p10unit <= majorMaxSpacing:
            majorScaleFactor = 10
            p10unit *= 10.
        else:
            for majorScaleFactor in (50, 20, 10):
                if majorScaleFactor * p10unit <= majorMaxSpacing:
                    break
        
        majorInterval = majorScaleFactor * p10unit
        levels = [
            (majorInterval, 0),
        ]

        if self.style['maxTickLevel'] >= 1:
            # Modified this section to choose a smaller interval for minor ticks
            if majorScaleFactor == 10:
                minorScaleFactor = 5  # Half of major interval for minor ticks
            elif majorScaleFactor == 20:
                minorScaleFactor = 10  # Half of major interval
            elif majorScaleFactor == 50:
                minorScaleFactor = 25  # Half of major interval
            else:
                minorScaleFactor = majorScaleFactor / 2  # Default to half
            
            minorInterval = minorScaleFactor * p10unit
            levels.append((minorInterval, 0))

        if self.style['maxTickLevel'] >= 2:
            # Keep the extra level logic but ensure it's smaller than minor interval
            if majorScaleFactor == 10:
                extraScaleFactor = 2  # Smaller than minor interval
            elif majorScaleFactor == 20:
                extraScaleFactor = 5  # Smaller than minor interval
            elif majorScaleFactor == 50:
                extraScaleFactor = 10  # Smaller than minor interval
            else:
                extraScaleFactor = minorScaleFactor / 2  # Default to half
            
            extraInterval = extraScaleFactor * p10unit
            if extraInterval < minorInterval:
                levels.append((extraInterval, 0))
        return levels

    def getWorldCoord(self, value):
        view = self.linkedView()
        if view is None:
            world_bounds = 0, 1
        else:
            world_bounds = view.state["worldRange"][self.axisDirection()]

        dif = self.range[1] - self.range[0]
        rmin, rmax = self.range
        if rmax == rmin:
            scale = 1
            offset = 0
            value
        else:
            scale = (world_bounds[1] - world_bounds[0]) / dif
            offset = world_bounds[0] - rmin * scale
        return scale * value + offset
    
    def labelString(self):
        return self.labelStyle["labelText"]
        
    def _linkToView_internal(self, view: GL3DViewBox):
        self.unlinkFromView()
        self._linkedView = weakref.ref(view)

        self._setViewBoxRange(view.state["worldRange"])
        
        if self.axisDirection() == 0:
            view.sigXRangeChanged.connect(self.linkedViewChanged)
        elif self.axisDirection() == 1:
            view.sigYRangeChanged.connect(self.linkedViewChanged)
        else:
            view.sigZRangeChanged.connect(self.linkedViewChanged)
        view.sigResized.connect(self.linkedViewChanged)
        if isinstance(view.parentWidget(), GLViewWidget):
            plotitem_widget = view.parentWidget()
                        
            if self.tickStyle["textAutoUpdateAnchor"] or self.labelStyle["labelAutoUpdateAnchor"]:
                    plotitem_widget.sigViewAngleChanged.connect(self.autoUpateAnchor)
                    self.autoUpateAnchor(plotitem_widget)
            if self.style["focus"]:
                plotitem_widget.sigViewAngleChanged.connect(self.updateFocusParameters)
           
    
    
    @QtCore.Slot(object)
    def updateFocusParameters(self, plotitem_widget: GL3DViewWidget):
        cameraPosition = plotitem_widget.cameraTuple()
        if self.axisDirection() == 2:
            angle = self.style["planeToCameraFn"](cameraPosition)
            #print(f"angle: {math.degrees(angle)}")
            
            
            
            
            return 

        current_angle = self.transformCondtions["azimuth"]["current"](cameraPosition)
        next_angle = self.transformCondtions["azimuth"]["next"](cameraPosition)
        """
        print("")
        print(self.axisDirection())

        print(f"current_angle: {math.degrees(current_angle)}")
        print(f"next_angle: {math.degrees(next_angle)}")
        """
        

        if self._axisGeometry.planeRectSize[2] == 0:
            if plotitem_widget.vb.worldRange():
                pass
        
        if current_angle > math.pi and next_angle < math.pi:
            self.flipAxis()
        
        
        
        
        if not 2 in (self.axisDirection(), self._axisGeometry.planeAxis):
            if cameraPosition[2] < self._axisGeometry.originPosition()[2]:
                if self.visible():
                    print(f"\nhiding:\n")
                    print(f"elevation: {plotitem_widget.opts['elevation']}")
                    print(f"angle: {math.degrees(current_angle)}")
                    print(f"cameraPosition: {cameraPosition}")

                    print(f"self._axisGeometry.originPosition()[2]: {self._axisGeometry.originPosition()[2]}")
                    self.hide()
            else:
                if not self.visible():
                    print(f"\nshowing:\n")
                    print(f"elevation: {plotitem_widget.opts['elevation']}")
                    print(f"angle: {math.degrees(current_angle)}")
                    print(f"cameraPosition: {cameraPosition}")

                    print(f"self._axisGeometry.originPosition()[2]: {self._axisGeometry.originPosition()[2]}")

                    self.show()
                
    def flipAxis(self):
        

        midPoint = self._axisGeometry.originPosition().copy() 
        midPoint[self._axisGeometry.planeAxis] += 0.5 * self._axisGeometry.planeRectSize[self._axisGeometry.planeAxis]
        
                
        
        tr = pg.Transform3D()
        tr.translate(*midPoint)

        scales = [1, 1, 1]
        scales[self._axisGeometry.planeAxis] *= -1
        
        tr.scale(*scales)
        tr.translate(*(-1 * midPoint))
        
        
        
        next = self.transformCondtions["azimuth"]["current"]
        current  = self.transformCondtions["azimuth"]["next"]
        
        
        self.transformCondtions["azimuth"]["next"]= next
        self.transformCondtions["azimuth"]["current"] = current
        
        
        self.setTransform(tr)
        

        """
        change[self.axisDirection()] = self._axisGeometry.planeRectSize[self.axisDirection()]
        self._GLItems["tickLineset"].translate(*change)
        """
        
    
        
    
    @QtCore.Slot(object)
    def autoUpateAnchor(self, plotitem_widget: GL3DViewWidget):
        
        cameraPosition = plotitem_widget.cameraTuple()
        
        if math.pi > self.style["planeToCameraFn"](cameraPosition):
            newAnchor = (1, self.style["valuesAnchor"][1])
        else:
            newAnchor = (0, self.style["valuesAnchor"][1])
        
        if self.labelStyle["labelAutoUpdateAnchor"]:
            self._GLItems["label"].setAnchor(newAnchor)
            
        if self.tickStyle["textAutoUpdateAnchor"]:
            if newAnchor != self.style["valuesAnchor"]:
                self.style["valuesAnchor"]=newAnchor
                for text_item_container in self._GLItems["text"]:
                    for text_item in text_item_container:
                        text_item.setAnchor(newAnchor)
                        
        if self.axisDirection() == 0:
            for text_item_container in self._GLItems["text"]:
                for text_item in text_item_container:
                    print("")
                    #print(cameraPosition)
                    #print(text_item.anchor())
                    print(self._axisGeometry.originPosition())
                    print(self.mapToParent(self._axisGeometry.originPosition()))
                    print(self.mapFromParent(self._axisGeometry.originPosition()))


    def linkToView(self, view):
        self._linkToView_internal(view)

    def unlinkFromView(self):
        oldView = self.linkedView()
        self._linkedView = None
        if oldView is not None:
            oldView.sigResized.disconnect(self.linkedViewChanged)
            if self.axisDirection() == 0:
                oldView.sigXRangeChanged.disconnect(self.linkedViewChanged)
            elif self.axisDirection() == 1:
                oldView.sigYRangeChanged.disconnect(self.linkedViewChanged)
            else:
                oldView.sigZRangeChanged.disconnect(self.linkedViewChanged)

    def linkedView(self) -> GL3DViewBox:
        return None if self._linkedView is None else self._linkedView()



    def setTickFont(self, font: QtGui.QFont | None):
        self.style['tickFont'] = font
        self.picture = None
        #self.prepareGeometryChange()
        self.update()


    def _buildOptimizers(self):
        self.style["textAnchorVar"] = tuple([int(val / abs(val)) if idx != self._axisGeometry.axis and val != 0 else 0 for idx, val in enumerate(self._axisGeometry.planeRectSize[:2])])
        planeRectSize = self._axisGeometry.planeRectSize.copy()[:2]
        textPositionOrigin = self._axisGeometry.tickTextPosition()
        planeLength = [int(val) if idx == self._axisGeometry.planeAxis else 0 for idx, val in enumerate(planeRectSize)]
        
        if planeLength[0] > 0 and planeLength[1] == 0:
            anchorCondition = lambda camera: (-1 * math.atan2(camera[1] - textPositionOrigin[1], camera[0] - textPositionOrigin[0])) % (2 * math.pi)
        elif planeLength[0] < 0 and planeLength[1] == 0:
            anchorCondition = lambda camera: (-1 * math.atan2(-(camera[1] - textPositionOrigin[1]), -(camera[0] - textPositionOrigin[0]))) % (2 * math.pi) 
        elif planeLength[0] == 0 and planeLength[1] > 0:
            anchorCondition = lambda camera: (-1 * math.atan2(-(camera[0] - textPositionOrigin[0]), camera[1] - textPositionOrigin[1])) % (2 * math.pi)
        elif planeLength[0] == 0 and planeLength[1] < 0:
            anchorCondition = lambda camera: (-1 * math.atan2(camera[0] - textPositionOrigin[0], -(camera[1] - textPositionOrigin[1]))) % (2 * math.pi)
        
        self.style["planeToCameraFn"] = anchorCondition
        
        if self.style["focus"]:
            if self.axisDirection() != 2:
            
                self.transformCondtions["azimuth"] = {"current" : None,
                                                       "next" : None}            
                def create_anchor_condition(plane_rect_size, axis_origin_position, axis_direction):
                    axis_length = [int(val) if idx == axis_direction else 0 for idx, val in enumerate(plane_rect_size)]
                    print(axis_length)
                    
                    if axis_length[0] > 0 and axis_length[1] == 0:
                        return lambda camera: (-1 * math.atan2(camera[1] - axis_origin_position[1], camera[0] - axis_origin_position[0])) % (2 * math.pi)
                    elif axis_length[0] < 0 and axis_length[1] == 0:
                        return lambda camera: (-1 * math.atan2(-(camera[1] - axis_origin_position[1]), -(camera[0] - axis_origin_position[0]))) % (2 * math.pi)
                    elif axis_length[0] == 0 and axis_length[1] > 0:
                        return lambda camera: (-1 * math.atan2(-(camera[0] - axis_origin_position[0]), camera[1] - axis_origin_position[1])) % (2 * math.pi)
                    elif axis_length[0] == 0 and axis_length[1] < 0:
                        return lambda camera: (-1 * math.atan2(camera[0] - axis_origin_position[0], -(camera[1] - axis_origin_position[1]))) % (2 * math.pi)
                
                plane_rect_size = self._axisGeometry.planeRectSize.copy()
                axis_origin_position = self._axisGeometry.originPosition().copy()
                self.transformCondtions["azimuth"]["current"] = create_anchor_condition(plane_rect_size, axis_origin_position, self.axisDirection())
                plane_rect_size = self._axisGeometry.planeRectSize.copy()
                axis_origin_position = self._axisGeometry.originPosition().copy() + plane_rect_size
                plane_rect_size*= -1
                self.transformCondtions["azimuth"]["next"] = create_anchor_condition(plane_rect_size, axis_origin_position, self.axisDirection())  
            else:
                plane_rect_size = self._axisGeometry.planeRectSize.copy()
                axis_origin_position = self._axisGeometry.originPosition().copy()
                axis_length = [int(val) if idx == self._axisGeometry.planeAxis else 0 for idx, val in enumerate(plane_rect_size)]
                print(axis_length)

    @classmethod
    def createFocusSet(cls, viewRange, **kwargs):
        pass


from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from pyvol_terminal.gl_3D_graphing.graphics_items.GL3DGraphicsItems import GlPlotItemMixin
    from pyvol_terminal.gl_3D_graphing.widgets.view_box import ViewBox, AxisNormaliser
    
    from uuid import uuid4 

import numpy as np
from dataclasses import dataclass, field
from pyqtgraph import opengl




def get_item_data(item):
    if isinstance(item, opengl.GLScatterPlotItem):
        return item.pos
    elif isinstance(item, opengl.GLMeshItem):
        return item.opts['meshdata'].vertexes()
    elif isinstance(item, opengl.GLLinePlotItem):
        return item.pos
    elif isinstance(item, opengl.GLBarGraphItem):
        return item.points()
    return None

def compute_scene_bbox(items):
    min_vals, max_vals = None, None
    
    for item in items:
        data = get_item_data(item)
        if data is None or len(data) == 0:
            continue
            
        if not isinstance(data, np.ndarray):
            data = np.array(data)
            
        if data.ndim == 1:
            data = data[np.newaxis, :]
        if data.shape[1] == 2:
            data = np.hstack([data, np.zeros((len(data), 1))])
            
        item_min = np.nanmin(data, axis=0)
        item_max = np.nanmax(data, axis=0)
        
        if min_vals is None:
            min_vals, max_vals = item_min, item_max
        else:
            min_vals = np.minimum(min_vals, item_min)
            max_vals = np.maximum(max_vals, item_max)
    
    if min_vals is None:
        return np.array([0, 0, 0]), np.array([1, 1, 1])
    
    return min_vals, max_vals







def normalise_setData_pos(item: GlPlotItemMixin,
                          vb: ViewBox,
                          ignoreBounds: bool,
                          *args,
                          **kwargs
                          ) -> np.ndarray:
    unnorm_pos = kwargs["pos"]
    if isinstance(unnorm_pos, tuple):
        unnorm_pos = np.array(unnorm_pos)
    item.unnorm_pos = unnorm_pos.copy()
    if not ignoreBounds:
        for idx, axis in enumerate("xyz"):
            vb._preNormalisation(item, axis, unnorm_pos[:, idx])
    pos=kwargs["pos"]
    _init_shape = np.shape(pos)
    x = vb.axis_normaliser_container["x"].normalise(pos[:, 0])
    y = vb.axis_normaliser_container["y"].normalise(pos[:, 1])
    z = vb.axis_normaliser_container["z"].normalise(pos[:, 2])
    kwargs["pos"] = np.column_stack((x , y , z)).reshape(_init_shape)
    return item.original_setData(*args, **kwargs)

def normalise_setData_xyz(item: GlPlotItemMixin,
                          vb: ViewBox,
                          ignoreBounds: bool,
                          *args,
                          **kwargs
                          ) -> Tuple[np.ndarray, ...]:
    new_args = list(args)
    if len(new_args) >= 3:
        for idx, axis in enumerate("xyz"):
            setattr(item, f"unnorm_{axis}", args[idx]) 
            vb._preNormalisation(item, axis, args[idx])
            #item.unnorm_pos[:, idx] = args[idx]
            
        for idx, axis in enumerate("xyz"):
            new_args[idx] = vb.axis_normaliser_container[axis].normalise(args[idx])
    else:
        for idx, axis in enumerate("xyz"):
            if axis in kwargs:
                setattr(item, f"unnorm_{axis}", kwargs[axis]) 
                #item.unnorm_pos[:, idx] = kwargs[axis]
                vb._preNormalisation(item, axis, kwargs[axis])
        for idx, axis in enumerate("xyz"):
            if axis in kwargs:
                kwargs[axis] = vb.axis_normaliser_container[axis].normalise(kwargs[axis])
    return item.original_setData(*new_args, **kwargs)

def renormalise_data_pos(item, plot_view):
    pos=item.unnorm_pos.copy()
    if plot_view.axis=="x":
        idx=0
    elif plot_view.axis=="y":
        idx=1
    else:
        idx=2
    if len(np.shape(pos)) == 1:
        if isinstance(pos, tuple):
            pos = np.array(pos)
        pos[idx] = plot_view.normalise(pos[idx])
    else:
        pos[:, idx] = plot_view.normalise(pos[:,idx])
    item.original_setData(pos=pos) 
    
def filter_data_pos(item, plot_view: AxisNormaliser):
    pos=item.unnorm_pos.copy()
    if plot_view.axis=="x":
        idx=0
    elif plot_view.axis=="y":
        idx=1
    else:
        idx=2
    if len(np.shape(pos)) == 1:
        if isinstance(pos, tuple):
            pos = np.array(pos)
        mask = plot_view.mask_from_data(pos[idx])
        pos[idx] = plot_view.normalise(pos[idx])
    else:
        mask = plot_view.mask_from_data(pos[idx])
        pos[:, idx] = plot_view.normalise(pos[:,idx])
    pos = pos[mask]
    item.original_setData(pos=pos) 

def filter_data_xyz(item: GlPlotItemMixin,
                    ax: str,
                    view_box: ViewBox
                    ):
    values = getattr(item, f"_all_{ax}").copy()
    plot_view = view_box.axis_normaliser_container[ax]
    values_normalised=plot_view.normalise(values)
    mask = plot_view.mask_from_data(values)
    values_normalised[~mask]=np.nan
    values_f = values.copy()
    values_f[~mask]=np.nan
    kwargs={plot_view.axis : values_normalised}

    if plot_view.axis!="z":
        if plot_view.axis=="x":
            
            values_z = item._all_z.copy()
            values_z[~mask, :]=np.nan
            values_z_normalised = view_box.axis_normaliser_container["z"].normalise(values_z)
        else:
            values_z = item._all_z.copy()
            values_z[:, ~mask]=np.nan
            values_z_normalised = view_box.axis_normaliser_container["z"].normalise(values_z)
        print(f"values_z: {(np.nanmin(values_z).round(3), np.nanmax(values_z).round(3))}")

        kwargs["z"] = values_z_normalised
    item.original_setData(**kwargs) 

def filter_data_xyz2(item, plot_view: AxisNormaliser):
    values = getattr(item, f"_all_{plot_view.axis}")
    new_values=plot_view.normalise(values)
    mask = plot_view.mask_from_data(values)
    new_values[~mask]=np.nan
    
    kwargs={plot_view.axis : new_values}

    if plot_view.axis!="z":

        if plot_view.axis=="x":
            
            #values_z = getattr(item, f"unnorm_z")
            values_z = item._all_z
            values_z[~mask, :]=np.nan
        else:
            #values_z = getattr(item, f"unnorm_z")
            values_z = item._all_z

            values_z[:, ~mask]=np.nan
        kwargs["z"] = values_z
    item.original_setData(**kwargs) 


def renormalise_data_xyz(item, plot_view):
    values = getattr(item, f"unnorm_{plot_view.axis}")
    new_values=plot_view.normalise(values)
    kwargs={plot_view.axis : new_values}
    item.original_setData(**kwargs) 




























    def filter_on_view222(self,
                       ax: str,
                       view_box
                       ):
        
        values = self._all_pos_vect[ax]
        plot_view = view_box.plot_views[ax]
        
        mask = plot_view.mask_from_data(values)
        values_in_view = values[~mask]
        self._in_view_vect[ax] = values_in_view
        
        values_onview_normalised = plot_view.normalise(values_in_view)
        kwargs = {self._ax_idx_str_map[ax] : values_onview_normalised}
        if ax != 2:
            values_z = self._all_pos_vect[2]
            if plot_view.axis==0:
                
                zvalues_in_view = values_z[~mask, :]
            else:
                zvalues_in_view = values_z[:, ~mask]
            
            self._in_view_vect[2] = zvalues_in_view
            
            zvalues_onview_normalised = view_box.plot_views[2].normalise(zvalues_in_view)
            
            kwargs[self._ax_idx_str_map[ax]]=zvalues_onview_normalised

        super().setData(**kwargs) 


    def id(self):
        return self._id
    






























class ItemLimists22222:
    def __init__(self, axes):
        self.axes=axes
        self.items = {}
        self.heap_min = {ax:[] for ax in self.axes}
        self.heap_max = {ax:[] for ax in self.axes}
        self.min_sorted={ax:[] for ax in self.axes}
        self.max_sorted={ax:[] for ax in self.axes}
        self.min_items = {ax:{} for ax in self.axes}
        self.max_items = {ax:{} for ax in self.axes}
        
    def _calculate_limits(self, values):
        if np.any(~np.isnan(values)):
            min, max = np.nanmin(values), np.nanmax(values)
        else:
            min, max = np.inf, -np.inf 
        return [min, max]
    
    def _get_value_vectors(self, item):
        vals=[]
        if hasattr(item, "pos"):
            pos = item.pos
            pos = np.atleast_2d(pos)
            for i in range(pos.shape[1]):
                vals.append(pos[:,i])
        else:
            for axis in self.axes:
                vals.append(getattr(item, f"_{axis}"))
        return vals
    
    def _calculate_vector_limits(self, vectors):
        limits=[]
        for vector in vectors:
            limits.append(self._calculate_limits(vector))
        return limits

    def get_limit_values(self, axis):
        return self.min_sorted[axis][0], self.max_sorted[axis][-1]

    def addItem(self, item):
        vectors = self._get_value_vectors(item)
        vector_limits = self._calculate_vector_limits(vectors)
        self.items[item.id()] = {}
        for axis, limits in zip(self.axes, vector_limits):
            self.items[item.id()][axis] = limits
            self.min_items[axis][item.id()]=limits[0]
            self.max_items[axis][item.id()]=limits[1]
            self._sort_limits(axis)

    def _sort_limits(self, axis):
        self.min_sorted[axis]=list(self.min_items[axis].values())
        self.min_sorted[axis].sort()
        self.max_sorted[axis]=list(self.max_items[axis].values())
        self.max_sorted[axis].sort()
    
    def update_item_limits(self, item, axis, values):
        min_val, max_val = self._calculate_limits(values)
        self.min_items[axis][item.id()]=min_val
        self.max_items[axis][item.id()]=max_val
        self._sort_limits(axis)

    def removeItem(self, internal_id):
        if internal_id in self.items:
            del self.items[internal_id]


from PySide6 import QtCore, QtGui, QtWidgets
from pyvol_terminal.settings import utils as settings_utils



class Settings(QtWidgets.QWidget):
    def __init__(self, toggle_slots_map, parent=None):
        self.dimension_actions={}
        self.title_axis_dict = {"Money" : "x", "Expiry" : "y", "Volatility" : "z"}

        super().__init__(parent)
        self.create_settings_objects(toggle_slots_map)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        
    def create_settings_objects(self, toggle_slots_map):
        self.setStyleSheet("background-color: white;")
        self.setLayout(QtWidgets.QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0 ,0)
        
        self.prev_button=1         
        
        self.create_change_dimensions_menu("Change Dimensions", toggle_slots_map["Change Dimensions"])
        self.create_toggle_buttons22222("Toggle Price Types", toggle_slots_map["Toggle Price Types"], ["bid", "ask", "mid"])
        
    def create_change_dimensions_menu(self, axis_direction, *arg):
        tool_button = QtWidgets.QToolButton(self)
        tool_button.setText("Change Dimensions")
        tool_button.setStyleSheet(settings_utils.get_settings_stylesheets("QToolButton"))

        tool_button.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        main_menu = QtWidgets.QMenu(self)
        main_menu.setStyleSheet(settings_utils.get_settings_stylesheets("QMenu"))

        def _add_submenu(title, options, axis):
            submenu = QtWidgets.QMenu(title, main_menu)
            main_menu.addMenu(submenu)

            for option in options:
                action = submenu.addAction(option)
                action.triggered.connect(lambda checked=False, o=option: self.widget_data_display.switch_axis(axis, o))
        
        if axis_direction == 0:
            _add_submenu("Money", ["Strike", "Delta", "Moneyness", "Log-Moneyness", "Standardised-Moneyness"], "x")
        else:
            _add_submenu("Expiry", ["Expiry", "Years"], "y")
        
        _add_submenu("Volatility", ["Implied Volatility", "Implied Volatility (%)", "Total Volatility"], "y")

        tool_button.setMenu(main_menu)
        tool_button.setPopupMode(QtWidgets.QToolButton.InstantPopup)

        column_layout = QtWidgets.QVBoxLayout()
        column_layout.addWidget(tool_button)
        self.layout().addLayout(column_layout)
    
    def create_toggle_buttons(self, title, trigger_functions, option_list):
        
        btn_group=QtWidgets.QButtonGroup()
        btn_layout=QtWidgets.QVBoxLayout()
        main_button = QtWidgets.QPushButton(self)
        main_button.setText(title)
        main_button.setStyleSheet(settings_utils.get_settings_stylesheets("QPushButton"))
        main_button.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)

        menu = QtWidgets.QMenu(self)
        menu.setStyleSheet(settings_utils.get_settings_stylesheets("QMenu"))

        buttons = []

        for idx, option in enumerate(option_list):
            btn = QtWidgets.QPushButton(option)
            btn.setCheckable(True)
            btn.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
            btn.setStyleSheet(settings_utils.get_settings_stylesheets("QPushButton"))
            buttons.append(btn)
            
            def on_click(btn=btn, opt=option):
                trigger_functions(opt)

            btn.clicked.connect(lambda _, b=btn, opt=option: on_click(b, opt))

            widget_action = QtWidgets.QWidgetAction(menu)
            widget_action.setDefaultWidget(btn)
            menu.addAction(widget_action)

            if option in "mid":
                btn.setChecked(True)
                #trigger_functions(option)

        main_button.clicked.connect(lambda: menu.exec_(main_button.mapToGlobal(QtCore.QPoint(0, main_button.height()))))

        column_layout = QtWidgets.QVBoxLayout()
        column_layout.addWidget(main_button)
        self.layout().addLayout(column_layout)    
             
    def create_toggle_buttons22222(self, title, trigger_functions, option_list):
        
        main_button = QtWidgets.QPushButton(self)
        main_button.setText(title)
        main_button.setStyleSheet(settings_utils.get_settings_stylesheets("QPushButton"))
        main_button.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)

        menu = QtWidgets.QMenu(self)
        menu.setStyleSheet(settings_utils.get_settings_stylesheets("QMenu"))

        buttons = []

        for idx, option in enumerate(option_list):
            btn = QtWidgets.QPushButton(option)
            btn.setCheckable(True)
            btn.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
            btn.setStyleSheet(settings_utils.get_settings_stylesheets("QPushButton"))
            buttons.append(btn)
            
            def on_click(btn=btn, opt=option):
                trigger_functions(opt)

            btn.clicked.connect(lambda _, b=btn, opt=option: on_click(b, opt))

            widget_action = QtWidgets.QWidgetAction(menu)
            widget_action.setDefaultWidget(btn)
            menu.addAction(widget_action)

            if option in "mid":
                btn.setChecked(True)
                #trigger_functions(option)

        main_button.clicked.connect(lambda: menu.exec_(main_button.mapToGlobal(QtCore.QPoint(0, main_button.height()))))

        column_layout = QtWidgets.QVBoxLayout()
        column_layout.addWidget(main_button)
        self.layout().addLayout(column_layout)    
             

             

class ToggleButton(QtWidgets.QPushButton):
    def __init__(self, text, value, widget_data_display, parent=None):
        super().__init__(text, parent)
        self.setCheckable(True)
        self.widget_data_display=widget_data_display
        
        #self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        self.setStyleSheet("""
                            QPushButton {
                                border: 1px solid #8f8f91;
                                border-radius: 4px;
                                color: black;
                                padding: 5px;
                                background-color: white;
                                text-align: left;
                            }
                            QPushButton:checked {
                                background-color: #0078d7;
                                color: black;
                            }
                            """)
        self.setIconSize(QtCore.QSize(16, 16))
        self.toggled.connect(lambda checked, val=value: self.update_icon(checked, val))

    def update_icon(self, checked, value):
        
        if checked:
            pixmap = QtGui.QPixmap(16, 16)
            pixmap.fill(QtCore.Qt.transparent)
            painter = QtGui.QPainter(pixmap)
            painter.setPen(QtGui.QColor("white"))
            painter.drawLine(4, 8, 7, 12)
            painter.drawLine(7, 12, 12, 4)
            painter.end()
            self.setIcon(QtGui.QIcon(pixmap))
            self.adjustSize()  
            self.widget_data_display.add_line(value)
        else:
            self.setIcon(QtGui.QIcon())
            self.adjustSize()  
            self.widget_data_display.remove_line(value)


class ScrollableToggleButtons(QtWidgets.QWidget):
    def __init__(self, slot, domain_vec, tick_engine):  
        super().__init__()
        self.slot = slot
        self.domain_vec=domain_vec
        self.tick_engine = tick_engine
        self.setLayout(QtWidgets.QVBoxLayout(self))
        self.layout().setAlignment(QtCore.Qt.AlignTop)
        self.create_buttons()
        self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        
    def create_buttons(self):
        self.button_group = QtWidgets.QButtonGroup() 
        self.button_group.setExclusive(False)
        self.button_layout=QtWidgets.QVBoxLayout()
        
        for idx, value in enumerate(self.domain_vec):
            text = self.tick_engine.function([value])[0]
            
            button = ToggleButton(str(text), value, self.slot)
            

            self.button_group.addButton(button, idx)
            self.layout().addWidget(button)            
                
        self.update_min_width()
        

    def update_min_width(self):
        return 
        max_width = 0
        for btn in self.buttons:
            was_checked = btn.isChecked()
            btn.click()
            btn.updateGeometry()
            width = btn.width()
            if width > max_width:
                max_width = width
            btn.click()
        
        self.setMinimumWidth(max_width)
        self.setFixedWidth(max_width)



from __future__ import annotations 
from typing import List, Optional, Tuple, Union, Dict, TYPE_CHECKING, Callable
if TYPE_CHECKING:
    from .GL3DViewBox import GL3DViewBox
    from enum import Flag 

import weakref
from pyqtgraph.Qt import QtCore, QtGui, QtWidgets, isQObjectAlive
import operator
from functools import reduce
from abc import ABCMeta, abstractmethod
import numpy as np
from ..import meta
from pyqtgraph import Transform3D
from pyqtgraph import functions as fn
import abc
import traceback
import inspect
import warnings
from pyqtgraph.Qt import QT_LIB
from PySide6.QtWidgets import QGraphicsItem
from PySide6 import QtCore
from OpenGL import GL
from functools import wraps

from enum import Enum, EnumType, Flag, IntFlag

from pyqtgraph.Qt import QtCore, QtGui

GLOptions = {
    'opaque': {
        GL.GL_DEPTH_TEST: True,
        GL.GL_BLEND: False,
        GL.GL_CULL_FACE: False,
    },
    'translucent': {
        GL.GL_DEPTH_TEST: True,
        GL.GL_BLEND: True,
        GL.GL_CULL_FACE: False,
        'glBlendFuncSeparate': (GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA,
                                GL.GL_ONE, GL.GL_ONE_MINUS_SRC_ALPHA),
    },
    'additive': {
        GL.GL_DEPTH_TEST: False,
        GL.GL_BLEND: True,
        GL.GL_CULL_FACE: False,
        'glBlendFunc': (GL.GL_SRC_ALPHA, GL.GL_ONE),
    },
}    




qt_enums = [QGraphicsItem.GraphicsItemChange.ItemPositionChange,
            QGraphicsItem.GraphicsItemChange.ItemPositionHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemVisibleChange, 
            QGraphicsItem.GraphicsItemChange.ItemVisibleHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemSceneChange, 
            QGraphicsItem.GraphicsItemChange.ItemTransformChange,
            QGraphicsItem.GraphicsItemChange.ItemTransformHasChanged,
            QGraphicsItem.GraphicsItemChange.ItemSceneHasChanged, 
            QGraphicsItem.GraphicsItemChange.ItemParentChange,
            QGraphicsItem.GraphicsItemChange.ItemParentHasChanged,
            QGraphicsItem.GraphicsItemChange.ItemChildAddedChange,
            QGraphicsItem.GraphicsItemChange.ItemChildRemovedChange,
            QGraphicsItem.GraphicsItemChange.ItemFlagsChange,
            QGraphicsItem.GraphicsItemChange.ItemFlagsHaveChanged
            ]
    

class GLGraphicsEnumMetaclass(EnumType):
    def __new__(metacls, clsname, bases, clsdict):
        base_enums = clsdict.get('_base_enums', [])
        custom_members = clsdict.get('_custom_members', [])

        ignore_list = list(clsdict.get('_ignore_', []))
        ignore_list.extend(['_base_enums', '_custom_members'])
        clsdict['_ignore_'] = ignore_list

        values = [member.value for member in base_enums]
        max_value = max(values) if values else 0

        for member in base_enums:
            clsdict[member.name] = member.value

        shifts = [1, 2]
        for idx, (name, _) in enumerate(custom_members):
            shift = shifts[idx] if idx < len(shifts) else (idx + 1)
            value = max_value << shift
            clsdict[name] = value

        return super().__new__(metacls, clsname, bases, clsdict)
    
    def __call__(cls, value):
        return super().__call__(value)


qt_flags = [QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations,
            QGraphicsItem.GraphicsItemFlag.ItemHasNoContents,
            QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges,
            
            ]


class MetaFlag(type(IntFlag)):
    def __call__(cls, value=0):
        return super().__call__(value)


class AbstractGLGraphicsItem(QtCore.QObject, metaclass=meta.QABCMeta):
    
    _nextId = 0
    
        
    class GLGraphicsItemChange(Enum, metaclass=GLGraphicsEnumMetaclass):
        _ignore_ = ['_base_enums', '_custom_members'] 
        _base_enums = qt_enums
        _custom_members = [('ItemViewChange', None), ('ItemViewHasChanged', None)]

    class GLGraphicsItemFlag(IntFlag, metaclass=MetaFlag):
        NoFlag = 0
        ItemIgnoresTransformations = QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations.value
        ItemHasNoContents = QGraphicsItem.GraphicsItemFlag.ItemHasNoContents.value
        ItemSendsGeometryChanges = QGraphicsItem.GraphicsItemFlag.ItemSendsGeometryChanges.value


        def __repr__(self):
            if self.value == 0:
                return f"{self.__class__.__name__}({self.value})"
            return super().__repr__()

    
    wrapped_methods = ["setData", "paint"]
    
    



    def __init__(self, parentItem: 'AbstractGLGraphicsItem' = None, **kwargs):
        super().__init__()
        self._id = AbstractGLGraphicsItem._nextId
        AbstractGLGraphicsItem._nextId += 1
        self.__blockUpdates: bool=False
        self.blockUpdates(False)
        self.__cachedView=None

        self.__parent: AbstractGLGraphicsItem | None = None
        self.__view = None
        self.__children: list[AbstractGLGraphicsItem] = list()
        self.__transform = Transform3D()
        self.__visible = True
        self.__flags = self.GLGraphicsItemFlag()
        
        self.setFlag(self.GLGraphicsItemFlag.ItemSendsGeometryChanges)
        (f"\nAbstract__init__: {parentItem}")
        self.__initialized = False
        self.setParentItem(parentItem)
        self.setDepthValue(0)
        self.__glOpts = {}

                
    def blockUpdates(self, flag):
        print(f"\nblockUpdates: {flag}")
        self.__blockUpdates=flag

    def setFlag(self, flag, enabled=True):
        if isinstance(flag, Flag):
            _ = self.itemChange(self.GLGraphicsItemChange.ItemFlagsChange, enabled)
            
            if enabled:
                new_flags = self.flags() | flag
            else:
                new_flags = self.flags() & ~flag
            
            if self.flags() != new_flags:
                self.__flags = new_flags
            _ = self.itemChange(self.GLGraphicsItemChange.ItemFlagsHaveChanged, enabled)
    
    def flags(self):    
        return self.__flags

    
    

    def setParentItem(self, item: 'AbstractGLGraphicsItem'):
        changed_enums=[]
        
        if not self.__parent is None:
            if item == self.__parent:
                warnings.warn(f"{item.__class__.__name__} is already a parentItem to {self.__class__.__name__}, ignoring...", UserWarning)
                return
            else:
                self.itemChange(self.GLGraphicsItemChange.ItemParentChange, item)
                self.itemChange(self.GLGraphicsItemChange.ItemChildRemovedChange, item)
                changed_enums.append((self.GLGraphicsItemChange.ItemParentHasChanged, item))
                self.__parent.__removeChild(self)
            
        if not item is None:    
            if len(changed_enums) == 0:
                self.itemChange(self.GLGraphicsItemChange.ItemParentChange, item)
                changed_enums.append((self.GLGraphicsItemChange.ItemParentHasChanged, item))
            changed_enums.append((self.GLGraphicsItemChange.ItemChildAddedChange, self))
            item.__addChild(self)
            
        if not self.__view is None:
            self.itemChange(self.GLGraphicsItemChange.ItemViewChange, item)
            changed_enums.append((self.GLGraphicsItemChange.ItemViewHasChanged, item))
            self.__view.removeItem(self)
        self.__parent = item
        self.__view = None
        
        for change, value in changed_enums:
            self.itemChange(change, value)
    
    
    def __addChild(self, childItem: 'AbstractGLGraphicsItem'):
        """Do NOT call this method or override it with setParentItem to the childItem. 
        Doing so will cause infinite recursion."""
        self.__children.append(childItem)

        
    def __removeChild(self, childItem: 'AbstractGLGraphicsItem'):
        if childItem in self.__children:
            self.__children.remove(childItem)

    def __paintHelper(self, *args, **kwargs):
        if QGraphicsItem.GraphicsItemFlag.ItemHasNoContents & self.flags():
            return 
        
        
    def visible(self):
        return self.__visible
        
    def itemChange(self, change, value):...

    
    
    def setGLOptions(self, opts):

        if isinstance(opts, str):
            opts = GLOptions[opts]
        self.__glOpts = opts.copy()
        self.update()
        
    def updateGLOptions(self, opts):
        """
        Modify the OpenGL state options to use immediately before drawing this item.
        *opts* must be a dictionary as specified by setGLOptions.
        Values may also be None, in which case the key will be ignored.
        """
        self.__glOpts.update(opts)
        
    
    def parentItem(self):
        """Return a this item's parent in the scenegraph hierarchy."""
        return self.__parent
        
        
    def _setView(self, v):
        self.__view = v
        
    def view(self):
        if self.__parent is None:
            # top level object
            return self.__view
        else:
            # recurse
            return self.__parent.view()
        
    def setDepthValue(self, value):

        self.__depthValue = value
        
    def depthValue(self):
        return self.__depthValue
    
    def __setTransformHelper(self, tr):
        self.itemChange(self.GLGraphicsItemChange.ItemTransformChange, tr)
        self.__transform = tr
        self.itemChange(self.GLGraphicsItemChange.ItemTransformHasChanged, tr)
        self.update()

        
    def setTransform(self, tr):
        if not isinstance(tr, Transform3D):
            tr = Transform3D(tr)
        if tr != self.transform():
            self.__setTransformHelper(tr)
        
    def resetTransform(self):
        if self.transform().isIdentity():
            return 
        self.__setTransformHelper(Transform3D())

        
    def applyTransform(self, tr, local):
        
        if local:
            newTransform = self.transform() * tr
        else:
            newTransform = tr * self.transform()
        if newTransform != self.transform():
            self.__setTransformHelper(newTransform)

            
    def transform(self) -> Transform3D:
        """Return this item's transform object."""
        return self.__transform
        
    def viewTransform(self):
        tr = self.__transform
        p = self
        while True:
            p = p.parentItem()
            if p is None:
                break
            tr = p.transform() * tr
        return Transform3D(tr)
        
    def translate(self, dx, dy, dz, local=False):
        tr = Transform3D()
        tr.translate(dx, dy, dz)
        self.applyTransform(tr, local=local)
        
    def rotate(self, angle, x, y, z, local=False):
        """
        Rotate the object around the axis specified by (x,y,z).
        *angle* is in degrees.
        
        """
        tr = Transform3D()
        tr.rotate(angle, x, y, z)
        self.applyTransform(tr, local=local)
    
    def scale(self, x, y, z, local=True):
        """
        Scale the object by (*dx*, *dy*, *dz*) in its local coordinate system.
        If *local* is False, then scale takes place in the parent's coordinates.
        """
        tr = Transform3D()
        tr.scale(x, y, z)
        self.applyTransform(tr, local=local)
    
    
    def hide(self):

        self.setVisible(False)
        
    def show(self):
        self.setVisible(True)
    
    
    def initialize(self):
        self.initializeGL()
        self.__initialized = True

    def isInitialized(self):
        return self.__initialized
    
    def initializeGL(self):

        pass
    
    def setupGLState(self):
        for k,v in self.__glOpts.items():
            if v is None:
                continue
            if isinstance(k, str):
                func = getattr(GL, k)
                func(*v)
            else:
                if v is True:
                    GL.glEnable(k)
                else:
                    GL.glDisable(k)
    
    def paint(self):
        self.setupGLState()
        
    def update(self):
        print("\nupdate\n")
        print(f"self.updateBlocked(): {self.updateBlocked()}")
        if self.updateBlocked():
            return

        v = self.view()
        print(f"v: {v}")
        if v is None:
            return
        v.update()
    
    def updateBlocked(self):
        return self.__blockUpdates
   

    def setVisible(self, value, explicitly):
        self.__setVisibleHelper(value, True, explicitly)

    def __setVisibleHelper(self, visible, update, explicitly):
        if explicitly:
            self.__explicitlyHidden = visible 
            
        if not self.parentItem() is None and visible and not self.parentItem().visible():
            return 

        if visible == self.visible():
            return 
            
        newVisible = self.itemChange(self.GLGraphicsItemChange.ItemVisibleChange, visible)
        if self.visible() == newVisible:
            return
        
        self.__visible = newVisible
        
        for child in self.childItems():
            if not visible or not child.__explicitlyHidden:
                child.__setVisibleHelper(visible, False, False)
        
        self.update()
        self.itemChange(self.GLGraphicsItemChange.ItemVisibleHasChanged, self.visible())
    
    def childItems(self) -> List['AbstractGLGraphicsItem']:
        return list(self.__children)
    
   
    def mapToParent(self, point):
        tr = self.transform()
        if tr is None:
            return point
        return tr.map(point)
        
    def mapFromParent(self, point):
        tr = self.transform()
        if tr is None:
            return point
        return tr.inverted()[0].map(point)
        
    def mapToView(self, point):
        tr = self.viewTransform()
        if tr is None:
            return point
        return tr.map(point)
        
    def mapFromView(self, point):
        tr = self.viewTransform()
        if tr is None:
            return point
        return tr.inverted()[0].map(point)

    def modelViewMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentModelView()

    def projectionMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentProjection()

    def mvpMatrix(self) -> QtGui.QMatrix4x4:
        if (view := self.view()) is None:
            return QtGui.QMatrix4x4()
        return view.currentProjection() * view.currentModelView()


    def _updateView(self):
        ## called to see whether this item has a new view to connect to
        ## NOTE: This is called from GraphicsObject.itemChange or GraphicsWidget.itemChange.

        if not hasattr(self, '_connectedView'):
            # Happens when Python is shutting down.
            return
        self.forgetViewBox()
        
        ## check for this item's current viewbox
        view = self.getViewBox()
        
        oldView = None
        if not self._connectedView is None:
            oldView = self._connectedView()
        
        if view is oldView:
            # "  already have view", view
            return

        ## disconnect from previous view
        if not oldView is None:
            for signal, slot in [('sigRangeChanged', self.viewRangeChanged),
                                 ('sigTransformChanged', self.viewTransformChanged)]:
                try:
                    getattr(oldView, signal).disconnect(slot)
                except (TypeError, AttributeError, RuntimeError):
                    # TypeError and RuntimeError are from pyqt and pyside, respectively
                    pass
            
            self._connectedView = None
        if not view is None:
            view.sigRangeChanged.connect(self.viewRangeChanged)
            view.sigTransformChanged.connect(self.viewTransformChanged)
            self._connectedView = weakref.ref(view)
            self.viewRangeChanged()
            self.viewTransformChanged()
        self._replaceView(oldView)        
        self.viewChanged(view, oldView)        

    def viewChanged(self, view, oldView):
        """Called when this item's view has changed
        (ie, the item has been added to or removed from a ViewBox)"""
        
    def viewRect(self):
        """Return the visible bounds of this item's ViewBox or GraphicsWidget,
        in the local coordinate system of the item."""
        if self.__cachedView is not None:
            return self.__cachedView

        # Note that in cases of early returns here, the view cache stays empty (None).
        view = self.getViewBox()
        if view is None:
            return None
        bounds = self.mapRectFromView(view.viewRect())
        if bounds is None:
            return None

        bounds = bounds.normalized()
        
        self.__cachedView = bounds
        
        ## nah.
        #for p in self.getBoundingParents():
            #bounds &= self.mapRectFromScene(p.sceneBoundingRect())

        return bounds
        
    def _replaceView(self, oldView, item=None):
        if item is None:
            item = self
        for child in item.childItems():
            if isinstance(child, 'AbstractGLGraphicsItem'):
                if child.getViewBox() is oldView:
                    child._updateView()
                        #self._replaceView(oldView, child)
            else:
                self._replaceView(oldView, child)
        

    def parentChanged(self):
        self.changeParent(self)


    def changeParent(self):
        self._updateView()

    
    @QtCore.Slot()
    def viewTransformChanged(self):
        self.__cachedView = None
        
    
        
    def informViewBoundsChanged(self):
        view = self.getViewBox()
        print(f"view: {view}")
        print(f)
        print(f"\ninformViewBoundsChanged: {view is not None and hasattr(view, 'implements') and view.implements('ViewBox')}")
        if view is not None and hasattr(view, 'implements') and view.implements('ViewBox'):
            view.itemBoundsChanged(self)
    

    def allChildItems(self, root=None):
        """Return list of the entire item tree descending from this item."""
        if root is None:
            root = self
        tree = []
        for ch in root.childItems():
            tree.append(ch)
            tree.extend(self.allChildItems(ch))
        return tree
    
    @QtCore.Slot(object, object)
    @QtCore.Slot(object, object, object)
    def viewRangeChanged(self, vb=None, ranges=None, changed=None):...


    def mapRectFromView(self, obj):
        vt = self.viewTransform()
        if vt is None:
            return None
        vt = fn.invertQTransform(vt)
        return vt.mapRect(obj)


    def getViewBox(self) -> GL3DViewBox:
        if self._viewBox is None:
            p = self
            while True:
                try:
                    p = p.parentItem()
                except RuntimeError:  ## sometimes happens as items are being removed from a scene and collected.
                    return None
                if p is None:
                    return None
                if hasattr(type(p), 'implements'):
                    if p.implements('ViewBox'):
                        self._viewBox = weakref.ref(p)
                        break
        
        return self._viewBox() if self._viewBox else None
    
    
    def forgetViewBox(self):
        self._viewBox = None




class AbstractGLGraphicsItem22(AbstractGLGraphicsItem, metaclass=meta.QABCMeta):
    def __init__(self, parentItem=None, **kwargs):
        self._blockUpdate: bool
        self.__inform_view_on_changes=True
        self.blockUpdates(False)
        self._viewBox: GL3DViewBox=None
        self._viewWidget=None
        self._connectedView=None
        super().__init__()
        self._connectedView=None

        self._GLGraphicsItem__parent = None 
        self._GLGraphicsItem__children = list() 
        self._GLGraphicsItem__view = None 
        self.__cachedView = None
        self.setParentItem(parentItem)


    def blockUpdates(self, flag):
        self._blockUpdate=flag
        
    @property
    def __parent(self):
        return self._GLGraphicsItem__parent
    
    @__parent.setter
    def __parent(self, parentItem):
        self._GLGraphicsItem__parent=parentItem
    
    @property
    def __view(self):
        return self._GLGraphicsItem__view
    
    @__view.setter
    def __view(self, view):
        self._GLGraphicsItem__view=view

    @property
    def __parent(self):
        return self._GLGraphicsItem__parent
    
    @__parent.setter
    def __parent(self, parentItem):
        self._GLGraphicsItem__parent=parentItem

    @property
    def __parent(self):
        return self._GLGraphicsItem__parent
    
    @__parent.setter
    def __parent(self, parentItem):
        self._GLGraphicsItem__parent=parentItem

    @property
    def __parent(self):
        return self._GLGraphicsItem__parent
    
    @__parent.setter
    def __parent(self, parentItem):
        self._GLGraphicsItem__parent=parentItem


    @property
    def __parent(self):
        return self._GLGraphicsItem__parent
    
    @__parent.setter
    def __parent(self, parentItem):
        self._GLGraphicsItem__parent=parentItem

    



    @property
    def __parent(self):
        return self._GLGraphicsItem__parent
    
    @__parent.setter
    def __parent(self, parentItem):
        
        self._GLGraphicsItem__parent=parentItem
        self.itemChange(self.GraphicsItemChange.ItemChildAddedChange, self)

    @abc.abstractmethod
    def _internal_update(self):...

    def update(self):
        ("update")
        if not self._blockUpdate:
            self._internal_update()
            
    def updatesBlocked(self):
        return self._blockUpdate

    @property
    def parent(self):
        """Getter for parent property"""
        return self._GLGraphicsItem__parent

    @parent.setter
    def parent(self, item: AbstractGLGraphicsItem):
        if self._GLGraphicsItem__parent is not None:
            self._GLGraphicsItem__parent._GLGraphicsItem__children.remove(self)
        if item is not None:
            item._GLGraphicsItem__children.append(self)
            item.itemChange(item.GraphicsItemChange.ItemChildAddedChange, self)
            

        if self._GLGraphicsItem__view is not None:
            self._GLGraphicsItem__view.removeItem(self)

        self._GLGraphicsItem__parent = item
        self.itemChange(self.GraphicsItemChange.ItemParentHasChanged, item)
        self._GLGraphicsItem__view = None

    def setParentItem(self, item):
        self.parent = item

    def boundingRect(self):
        return QtWidgets.QGraphicsItem.boundingRect(self)

    def childItems(self):
        return list(self._GLGraphicsItem__children)

    def getViewBox2(self) -> GL3DViewBox:
        return self._viewBox
    
    
    
    def getViewBox(self) -> GL3DViewBox:
        if self._viewBox is None:
            p = self
            while True:
                try:
                    p = p.parentItem()
                except RuntimeError:  ## sometimes happens as items are being removed from a scene and collected.
                    return None
                if p is None:
                    return None
                if hasattr(type(p), 'implements'):
                    if p.implements('ViewBox'):
                        self._viewBox = weakref.ref(p)
                        break
        
        return self._viewBox() if self._viewBox else None
    
    
        
    
    def parentChanged(self):
        self.changeParent(self)


    def changeParent(self):
        self._updateView()
        
    def forgetViewWidget(self):
        self._viewWidget = None
        
    def forgetViewBox(self):
        self._viewBox = None

    def _updateView2(self):
        ## called to see whether this item has a new view to connect to
        ## NOTE: This is called from GraphicsObject.itemChange or GraphicsWidget.itemChange.


        ## It is possible this item has moved to a different ViewBox or widget;
        ## clear out previously determined references to these.
        self.forgetViewBox()
        self.forgetViewWidget()
        
        ## check for this item's current viewbox or view widget
        view = self.getViewBox()
        #if view is None:
            ## "  no view"
            #return

        oldView = None
        if self._connectedView is not None:
            oldView = self._connectedView()
            
        if view is oldView:
            # "  already have view", view
            return

        if oldView is not None:
            Device = 'Device' if hasattr(oldView, 'sigDeviceRangeChanged') else ''
            for signal, slot in [(f'sig{Device}RangeChanged', self.viewRangeChanged),
                                 (f'sig{Device}TransformChanged', self.viewTransformChanged)]:
                try:
                    getattr(oldView, signal).disconnect(slot)
                except (TypeError, AttributeError, RuntimeError):
                    # TypeError and RuntimeError are from pyqt and pyside, respectively
                    pass
            
            self._connectedView = None

        if view is not None:
            # "connect:", self, view
            if hasattr(view, 'sigDeviceRangeChanged'):
                # connect signals from GraphicsView
                view.sigDeviceRangeChanged.connect(self.viewRangeChanged)
                view.sigDeviceTransformChanged.connect(self.viewTransformChanged)
            else:
                # connect signals from ViewBox
                view.sigRangeChanged.connect(self.viewRangeChanged)
                view.sigTransformChanged.connect(self.viewTransformChanged)
            self._connectedView = weakref.ref(view)
            self.viewRangeChanged()
            self.viewTransformChanged()
        
        ## inform children that their view might have changed
        self._replaceView(oldView)
        
        self.viewChanged(view, oldView)
        
            
    def _updateView(self):
        ## called to see whether this item has a new view to connect to
        ## NOTE: This is called from GraphicsObject.itemChange or GraphicsWidget.itemChange.

        if not hasattr(self, '_connectedView'):
            # Happens when Python is shutting down.
            return
        self.forgetViewBox()
        
        ## check for this item's current viewbox
        view = self.getViewBox()
        
        oldView = None
        if not self._connectedView is None:
            oldView = self._connectedView()
        
        if view is oldView:
            # "  already have view", view
            return

        ## disconnect from previous view
        if not oldView is None:
            for signal, slot in [('sigRangeChanged', self.viewRangeChanged),
                                 ('sigTransformChanged', self.viewTransformChanged)]:
                try:
                    getattr(oldView, signal).disconnect(slot)
                except (TypeError, AttributeError, RuntimeError):
                    # TypeError and RuntimeError are from pyqt and pyside, respectively
                    pass
            
            self._connectedView = None
        if not view is None:
            view.sigRangeChanged.connect(self.viewRangeChanged)
            view.sigTransformChanged.connect(self.viewTransformChanged)
            self._connectedView = weakref.ref(view)
            self.viewRangeChanged()
            self.viewTransformChanged()
        self._replaceView(oldView)        
        self.viewChanged(view, oldView)        

    def viewChanged(self, view, oldView):
        """Called when this item's view has changed
        (ie, the item has been added to or removed from a ViewBox)"""
        
    def viewRect(self):
        """Return the visible bounds of this item's ViewBox or GraphicsWidget,
        in the local coordinate system of the item."""
        if self.__cachedView is not None:
            return self.__cachedView

        # Note that in cases of early returns here, the view cache stays empty (None).
        view = self.getViewBox()
        if view is None:
            return None
        bounds = self.mapRectFromView(view.viewRect())
        if bounds is None:
            return None

        bounds = bounds.normalized()
        
        self.__cachedView = bounds
            
        ## nah.
        #for p in self.getBoundingParents():
            #bounds &= self.mapRectFromScene(p.sceneBoundingRect())

        return bounds
        
    def _replaceView(self, oldView, item=None):
        if item is None:
            item = self
        for child in item.childItems():
            if isinstance(child, AbstractGLGraphicsItem):
                if child.getViewBox() is oldView:
                    child._updateView()
                        #self._replaceView(oldView, child)
            else:
                self._replaceView(oldView, child)
        


    
    @QtCore.Slot()
    def viewTransformChanged(self):
        self.__cachedView = None
        
    
        
    def informViewBoundsChanged(self):
        view = self.getViewBox()
        
        if view is not None and hasattr(view, 'implements') and view.implements('ViewBox'):
            view.itemBoundsChanged(self)
    

    def allChildItems(self, root=None):
        """Return list of the entire item tree descending from this item."""
        if root is None:
            root = self
        tree = []
        for ch in root.childItems():
            tree.append(ch)
            tree.extend(self.allChildItems(ch))
        return tree
    
    @QtCore.Slot(object, object)
    @QtCore.Slot(object, object, object)
    def viewRangeChanged(self, vb=None, ranges=None, changed=None):...


    def mapRectFromView(self, obj):
        vt = self.viewTransform()
        if vt is None:
            return None
        vt = fn.invertQTransform(vt)
        return vt.mapRect(obj)






from PySide6 import QtWidgets, QtCore, QtGui

class OptionExpiryCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, text):
        super().__init__(text)
        self.setBackground(QtGui.QBrush(QtGui.QColor("orange")))
        self.setForeground(QtGui.QBrush(QtGui.QColor("white")))
        self.setTextAlignment(QtCore.Qt.AlignLeft)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)

class OptionMetricCellItem(QtWidgets.QTableWidgetItem):
    def __init__(self, text):
        super().__init__(text)
        self.setBackground(QtGui.QBrush(QtGui.QColor("black")))
        self.setForeground(QtGui.QBrush(QtGui.QColor("white")))
        self.setTextAlignment(QtCore.Qt.AlignRight)
        font = QtGui.QFont("Neue Haas Grotesk", 12)
        self.setFont(font)

class CustomDelegate(QtWidgets.QStyledItemDelegate):
    def __init__(self, table):
        super().__init__()
        self.table = table

    def paint(self, painter, option, index):
        item = self.table.item(index.row(), index.column())
        
        if item and item.background().color() == QtGui.QColor("orange"):
            # Fill background
            painter.fillRect(option.rect, QtGui.QColor("orange"))
            
            # Draw text manually with proper alignment and style
            painter.save()
            painter.setFont(item.font())
            painter.setPen(item.foreground().color())
            
            # Calculate text rectangle with padding
            text_rect = option.rect.adjusted(4, 0, -4, 0)  # 4px horizontal padding
            
            # Draw text with alignment
            painter.drawText(
                text_rect,
                item.textAlignment() | QtCore.Qt.AlignVCenter,
                item.text()
            )
            painter.restore()
        else:
            # Default painting for other items
            super().paint(painter, option, index)

app = QtWidgets.QApplication([])
window = QtWidgets.QMainWindow()
table = QtWidgets.QTableWidget(10, 4)

# Set delegate with table reference
table.setItemDelegate(CustomDelegate(table))


table.setStyleSheet("""
    QTableWidget {
        background-color: black;
        border-left: 2px solid #414141;
        border-right: 2px solid #414141;
    }
    QTableWidget::item {
        border: 0px;
        padding: 4px;
    }
    QTableWidget::item:selected {
        background-color: #414141;
    }
""")

# Configure table
table.setShowGrid(False)
#table.setFrameShape(QtWidgets.QFrame.NoFrame)
table.verticalHeader().setVisible(False)
table.horizontalHeader().setVisible(False)

# Fill the table
for row in range(10):
    for col in range(4):
        if row == 0:
            item = OptionExpiryCellItem(f"Expiry {col+1}")
        else:
            item = OptionMetricCellItem(f"{row}.{col}")
        table.setItem(row, col, item)

window.setCentralWidget(table)
window.resize(800, 600)
window.show()
app.exec()


from PySide6 import QtWidgets, QtCore
from . import utils


class MainSettings(QtWidgets.QWidget):
    def __init__(self, widget_main=None, interfaces=None):
        super().__init__()
        self.widget_main=widget_main
        self.idx_to_interface_name = {}
        
        self.starting_button=0
        self.prev_btn_idx=0

        for idx, interface_name in enumerate(interfaces):            
            self.idx_to_interface_name[idx] = interface_name
            
        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(0, 0, 0 ,0)
        self._create_interface(interfaces)
        self.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)
        
        
    @QtCore.Slot(QtWidgets.QPushButton)
    def on_button_clicked(self, button):
        btn_id = self.button_group.id(button)
        if btn_id == self.prev_btn_idx:
            return 
        else:
            self.prev_btn_idx = btn_id
        
        interface_name = self.idx_to_interface_name[btn_id]
        
        
        self.widget_main.switch_view(interface_name)
        
    
    def _create_interface(self, interface_widgets):        
        self.button_group = QtWidgets.QButtonGroup()
        self.button_group.setExclusive(True)

        self.buttons = []
        self.selection_labels = list(interface_widgets.keys())
        self.button_name_maps = {i : interface_name for i, interface_name in enumerate(self.selection_labels)}
        
        for i, label in enumerate(self.selection_labels):
            btn = QtWidgets.QPushButton(label)
            btn.setCheckable(True)
            btn.setStyleSheet(utils.get_settings_stylesheets("QPushButton"))
            self.buttons.append(btn)
            self.layout.addWidget(btn)
            self.button_group.addButton(btn, id=i) 
            btn.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)

        self.button_group.buttonClicked.connect(self.on_button_clicked)
        self.button_group.blockSignals(True)  
        self.buttons[self.starting_button].setChecked(True)     
        self.button_group.blockSignals(False)       
        



from PySide6 import QtCore, QtWidgets
import sys
import time

class Worker(QtCore.QObject):
    
    def __init__(self, name):
        super().__init__()
        self.name = name
       # self._thread = QtCore.QThread()
      #  self._
      #  self._thread.started.connect(self.start_thread)
        
    def start_thread(self):
        self.moveToThread(self._thread)
        self._thread.started.connect(self.work)
        self._thread.start()
    
    def work2(self):
        while True:
            print(f"Thead {self.name} working in thread {QtCore.QThread.currentThread()}...")
            time.sleep(1)
    
    def work(self):
        print(f"Thead {self.name} working in thread {QtCore.QThread.currentThread()}...")
        
    def start(self):
        print("starting worker")
        self._timer = QtCore.QTimer(self)
        self._timer.timeout.connect(self.work)
        self._timer.start(2000)
        
        
class Widget(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        self._workers=[]
        self.button = QtWidgets.QPushButton("Start Worker")
        self.button.clicked.connect(self.spawn_worker)
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.button)
        self.setLayout(layout)
        self._threads = []
        self.worker_count = 0
        
    def spawn_worker(self):
        self.worker_count += 1
        worker = Worker(f"Worker {self.worker_count}")
        thread = QtCore.QThread()
        self._threads.append(thread)
        
        worker.moveToThread(thread)
        
        thread.started.connect(worker.start)
        thread.start()
        
        #self._workers.append(worker)
        #worker.start_thread()

def main():
    app = QtWidgets.QApplication(sys.argv)
    widget = Widget()
    widget.show()
    sys.exit(app.exec())
    
if __name__ == "__main__":
    main()



def legend_colourmap_getter(legend_labels):
    
    list_of_colours = ["white", "red", "blue", "green", "yellow", "cyan", "magenta",
                        "orange", "purple", "pink", "brown",
                        "lightgray", "darkgray", "dimgray",
                        "lightred", "darkred",
                        "lightgreen", "darkgreen", "lime", "limegreen",
                        "lightblue", "darkblue", "navy",
                        "lightcyan", "darkcyan", "teal",
                        "lightmagenta", "darkmagenta", "fuchsia",
                        "lightyellow", "darkyellow", "gold", "goldenrod", "khaki",
                        "lightpink", "darkpink", "hotpink",
                        "lightsalmon", "darksalmon", "coral",
                        "lightgoldenrod", "darkgoldenrod",
                        ]      
    colourmap = {}
    for expiry, colour in zip(legend_labels, list_of_colours):
         colourmap[expiry] = colour
    return colourmap


class MetricRounder:
     def __init__(self, metric_type):
          self.metric_type=metric_type




import sys
import numpy as np
import pyqtgraph as pg
from pyqtgraph.Qt import QtWidgets, QtCore





class AutoPanDemo(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("AutoPan ON (left) vs OFF (right)")

        # Layout
        central_widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(central_widget)
        self.setCentralWidget(central_widget)

        # Left plot: autoPan ON
        self.plot_auto = pg.PlotWidget(title="autoPan = ON")
        self.plot_auto.plotItem.vb.setAutoPan(x=True)  # This is the key line
        self.curve_auto = self.plot_auto.plot(pen='y')
        layout.addWidget(self.plot_auto)

        # Right plot: autoPan OFF
        self.plot_manual = pg.PlotWidget(title="autoPan = OFF")
        # No autoPan here
        self.curve_manual = self.plot_manual.plot(pen='r')
        layout.addWidget(self.plot_manual)

        # Data
        self.x = []
        self.y = []
        self.counter = 0

        # Timer
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.update_data)
        self.timer.start(500)  # every 0.5 sec

    def update_data(self):
        self.counter += 1
        x_val = self.counter
        y_val = np.sin(x_val * 0.2)

        self.x.append(x_val)
        self.y.append(y_val)

        self.curve_auto.setData(self.x, self.y)
        self.curve_manual.setData(self.x, self.y)

        # Manually set range on the right plot to fix the view
        if self.counter == 1:
            self.plot_manual.setXRange(0, 20)  # fixed range; doesn't move
            
        plotitem = self.plot_auto.getPlotItem()
        
        x_range = plotitem.vb.state['limits']['xRange']
        x_lims = plotitem.vb.state['limits']['xLimits']
        print("")
        print(x_range)
        print(x_lims)
        
if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    window = AutoPanDemo()
    window.resize(1000, 400)
    window.show()
    sys.exit(app.exec())
